import pytest
import uuid
from datetime import datetime
from decimal import Decimal
from sqlalchemy_api_handler import ApiHandler
from sqlalchemy_api_handler.bases.errors import DateTimeCastError, \
                                                DecimalCastError, \
                                                EmptyFilterError, \
                                                ResourceNotFoundError, \
                                                UuidCastError
from sqlalchemy_api_handler.utils import dehumanize, \
                                         humanize, \
                                         NonDehumanizableId
from sqlalchemy_api_handler.serialization import as_dict

from tests.conftest import with_delete
from api.models.foo import Foo
from api.models.offer import Offer
from api.models.offerer import Offerer
from api.models.scope import ScopeType
from api.models.stock import Stock
from api.models.tag import Tag, TagType
from api.models.user import User
from api.models.user_offerer import UserOfferer
from api.models.time_interval import TimeInterval


time_interval = TimeInterval()
time_interval.start = datetime(2018, 1, 1, 10, 20, 30, 111000)
time_interval.end = datetime(2018, 2, 2, 5, 15, 25, 222000)
now = datetime.utcnow()

class ModifyTest:
    def test_user_string_fields_are_stripped_of_whitespace(self):
        # Given
        user_data = {
            'email': '   test@example.com',
            'firstName': 'John   ',
            'lastName': None,
            'postalCode': '   93100   ',
            'publicName': ''
        }

        # When
        user = User(**user_data)

        # Then
        assert user.email == 'test@example.com'
        assert user.firstName == 'John'
        assert user.lastName == None
        assert user.postalCode == '93100'
        assert user.publicName == ''

    def test_for_sql_integer_value_with_string_raises_decimal_cast_error(self):
        # Given
        test_object = Foo()
        data = {'integer_attribute': 'yolo'}

        # When
        with pytest.raises(DecimalCastError) as errors:
            test_object.modify(data)

        # Then
        assert errors.value.errors['integer_attribute'] == ["Invalid value for integer_attribute (integer): 'yolo'"]

    def test_for_sql_integer_value_with_str_12dot9_sets_attribute_to_12dot9(self):
        # Given
        test_object = Foo()
        data = {'integer_attribute': '12.9'}

        # When
        test_object.modify(data)

        # Then
        assert test_object.integer_attribute == Decimal('12.9')

    def test_for_sql_float_value_with_str_12dot9_sets_attribute_to_12dot9(self):
        # Given
        test_object = Foo()
        data = {'float_attribute': '12.9'}

        # When
        test_object.modify(data)

        # Then
        assert test_object.float_attribute == Decimal('12.9')

    def test_for_sql_integer_value_with_12dot9_sets_attribute_to_12dot9(self):
        # Given
        test_object = Foo()
        data = {'integer_attribute': 12}

        # When
        test_object.modify(data)

        # Then
        assert test_object.integer_attribute == 12

    def test_for_sql_float_value_with_12dot9_sets_attribute_to_12dot9(self):
        # Given
        test_object = Foo()
        data = {'float_attribute': 12.9}

        # When
        test_object.modify(data)

        # Then
        assert test_object.float_attribute == 12.9

    def test_for_valid_sql_uuid_value(self):
        # Given
        test_object = Foo()
        uuid_attribute = str(uuid.uuid4())
        data = {'uuid_attribute': uuid_attribute}

        # When
        test_object.modify(data)

        # Then
        assert test_object.uuid_attribute == uuid_attribute

    def test_for_valid_sql_uuid_value_with_key_finishing_by_Id(self):
        # Given
        test_object = Foo()
        uuid_id = str(uuid.uuid4())
        data = {'uuidId': uuid_id}

        # When
        test_object.modify(data)

        # Then
        assert test_object.uuidId == uuid_id

    def test_for_valid_sql_humanize_id_value_with_key_finishing_by_Id(self):
        # Given
        test_object = Foo()
        humanized_entity_id = "AE"
        data = {'entityId': humanized_entity_id}

        # When
        test_object.modify(data)

        # Then
        assert test_object.entityId == dehumanize(humanized_entity_id)


    def test_for_valid_sql_humanize_id_synonym_value_with_key_finishing_by_Id(self):
        # Given
        test_object = Foo()
        humanized_bar_id = "AE"
        dehumanized_bar_id = dehumanize(humanized_bar_id)
        data = {'barId': humanized_bar_id}

        # When
        test_object.modify(data)

        # Then
        assert test_object.bar_id == dehumanized_bar_id
        assert test_object.barId == dehumanized_bar_id


    def test_for_valid_relationship_dict(self):
        # Given
        test_object = Stock()
        offer_dict = {
            'name': 'foo',
            'type': 'bar'
        }
        stock_dict = {
            'offer': offer_dict,
            'price': 1
        }

        # When
        test_object.modify(stock_dict)

        # Then
        assert test_object.price == stock_dict['price']
        assert test_object.offer.name == offer_dict['name']


    def test_for_valid_relationship_dicts(self):
        # Given
        test_object = Offer()
        stock_dict1 = {
            'price': 1
        }
        stock_dict2 = {
            'price': 1
        }
        offer_dict = {
            'name': 'foo',
            'stocks': [stock_dict1, stock_dict2],
            'type': 'bar'
        }

        # When
        test_object.modify(offer_dict)

        # Then
        assert test_object.name == offer_dict['name']
        assert test_object.stocks[0].price == stock_dict1['price']
        assert test_object.stocks[1].price == stock_dict2['price']


    def test_for_sql_float_value_with_string_raises_decimal_cast_error(self):
        # Given
        test_object = Foo()
        data = {'float_attribute': 'yolo'}

        # When
        with pytest.raises(DecimalCastError) as errors:
            test_object.modify(data)

        # Then
        assert errors.value.errors['float_attribute'] == ["Invalid value for float_attribute (float): 'yolo'"]

    def test_for_sql_datetime_value_in_wrong_format_returns_400_and_affected_key_in_error(self):
        # Given
        test_object = Foo()
        data = {'date_attribute': {'date_attribute': None}}

        # When
        with pytest.raises(DateTimeCastError) as errors:
            test_object.modify(data)

        # Then
        assert errors.value.errors['date_attribute'] == [
            "Invalid value for date_attribute (datetime): {'date_attribute': None}"]

    def test_deserializes_datetimes(self):
        # Given
        raw_data = {'start': '2018-03-03T15:25:35.123Z', 'end': '2018-04-04T20:10:30.456Z'}

        # When
        time_interval.modify(raw_data)

        # Then
        assert time_interval.start == datetime(2018, 3, 3, 15, 25, 35, 123000)
        assert time_interval.end == datetime(2018, 4, 4, 20, 10, 30, 456000)

    def test_deserializes_datetimes_without_milliseconds(self):
        # Given
        raw_data = {'start': '2018-03-03T15:25:35', 'end': '2018-04-04T20:10:30'}

        # When
        time_interval.modify(raw_data)

        # Then
        assert time_interval.start == datetime(2018, 3, 3, 15, 25, 35)
        assert time_interval.end == datetime(2018, 4, 4, 20, 10, 30)

    def test_deserializes_datetimes_without_milliseconds_with_trailing_z(self):
        # Given
        raw_data = {'start': '2018-03-03T15:25:35Z', 'end': '2018-04-04T20:10:30Z'}

        # When
        time_interval.modify(raw_data)

        # Then
        assert time_interval.start == datetime(2018, 3, 3, 15, 25, 35)
        assert time_interval.end == datetime(2018, 4, 4, 20, 10, 30)

    def test_raises_type_error_if_raw_date_is_invalid(self):
        # Given
        raw_data = {'start': '2018-03-03T15:25:35.123Z', 'end': 'abcdef'}

        # When
        with pytest.raises(DateTimeCastError) as errors:
            time_interval.modify(raw_data)

        # Then
        assert errors.value.errors['end'] == ["Invalid value for end (datetime): 'abcdef'"]

    def test_raises_type_error_if_raw_uuid_is_invalid(self):
        # Given
        test_object = Foo()
        data = {'uuidId': 'foo'}

        # When
        with pytest.raises(UuidCastError) as errors:
            test_object.modify(data)

        # Then
        assert errors.value.errors['uuidId'] == [
            "Invalid value for uuidId (uuid): 'foo'"]

    def test_raises_type_error_if_raw_humanized_id_is_invalid(self):
        # Given
        test_object = Foo()
        data = {'entityId': '12R-..2foo'}

        # When
        with pytest.raises(NonDehumanizableId):
            test_object.modify(data)

    @with_delete
    def test_find_raise_empty_filter_error(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        with pytest.raises(EmptyFilterError) as errors:
            Offer.find({
                '__SEARCH_BY__': 'position',
                'name': 'fee',
                'type': 'bric'
            })

        # Then
        assert errors.value.errors['_filter_from'] == ["None of filters found among: position"]

    @with_delete
    def test_find_returns_none(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.find({
            '__SEARCH_BY__': 'name',
            'name': 'fee',
            'type': 'bric'
        })

        # Then
        assert offer2 is None

    @with_delete
    def test_find_returns_existing_offer(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.find({
            '__SEARCH_BY__': 'name',
            'name': 'foo'
        })

        # Then
        assert offer2.id == offer1.id
        assert offer2.name == offer1.name == 'foo'
        assert offer2.type == offer1.type == 'bar'

    @with_delete
    def test_find_or_create_returns_existing_offer(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.find_or_create({
            '__SEARCH_BY__': 'name',
            'name': 'foo'
        })

        # Then
        assert offer2.id == offer1.id
        assert offer2.name == offer1.name == 'foo'
        assert offer2.type == offer1.type == 'bar'

    @with_delete
    def test_find_or_create_returns_created_offer(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.find_or_create({
            '__SEARCH_BY__': 'name',
            'name': 'fee',
            'type': 'gold'
        })

        # Then
        assert offer2.id != offer1.id
        assert offer2.name == 'fee'
        assert offer2.type == 'gold'

    @with_delete
    def test_find_and_modify_returns_modified_offer(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.find_and_modify({
            '__SEARCH_BY__': 'name',
            'name': 'foo',
            'type': 'bric'
        })

        # Then
        assert offer2.id == offer1.id
        assert offer2.name == offer1.name == 'foo'
        assert offer2.type == 'bric'

    @with_delete
    def test_find_and_modify_raises_ressource_not_found_error(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        with pytest.raises(ResourceNotFoundError) as e:
            Offer.find_and_modify({
                '__SEARCH_BY__': 'name',
                'name': 'fee',
                'type': 'bric'
            })

        # Then
        assert e.value.errors['find_and_modify'] == ['No ressource found with {"name": "fee"} ']

    @with_delete
    def test_create_or_modify_returns_created_offer(self, app):
        # Given
        offer1 = Offer(name='foo', type='bar')
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.create_or_modify({
            '__SEARCH_BY__': 'name',
            'name': 'fee',
            'type': 'bric'
        })

        # Then
        assert offer2.id != offer1.id
        assert offer2.name == 'fee'
        assert offer2.type == 'bric'

    @with_delete
    def test_create_or_modify_returns_modified_offer(self, app):
        # Given
        offer1 = Offer(name="foo", type="bar")
        ApiHandler.save(offer1)

        # When
        offer2 = Offer.create_or_modify({
            '__SEARCH_BY__': 'name',
            'name': 'foo',
            'type': 'bric'
        })

        # Then
        assert offer2.id == offer1.id
        assert offer2.name == offer1.name == 'foo'
        assert offer2.type == 'bric'


    @with_delete
    def test_create_or_modify_returns_modified_offerer_search_by_id(self, app):
        # Given
        offerer1 = Offerer(name="foo")

        ApiHandler.save(offerer1)

        # When
        offerer2 = Offerer.create_or_modify({
            '__SEARCH_BY__': 'id',
            'id': humanize(offerer1.id),
            'name': 'fee'
        })

        # Then
        assert offerer2.id == offerer1.id
        assert offerer2.name == "fee"

    @with_delete
    def test_create_or_modify_returns_created_user_offerer_search_by_relationship_ids(self, app):
        # Given
        offerer = Offerer(name="foo")
        user = User(email="foo.marx@com", publicName="Foo Marx")
        ApiHandler.save(offerer, user)

        # When
        user_offerer = UserOfferer.create_or_modify({
            '__SEARCH_BY__': ['offererId', 'userId'],
            'rights': 'admin',
            'offererId': humanize(offerer.id),
            'userId': humanize(user.id)
        })

        # Then
        assert user_offerer.offererId == offerer.id
        assert user_offerer.rights == 'admin'
        assert user_offerer.userId == user.id


    @with_delete
    def test_create_or_modify_returns_modified_user_offerer_search_by_relationship_ids(self, app):
        # Given
        offerer = Offerer(name="foo")
        user = User(email="foo.marx@com", publicName="Foo Marx")
        user_offerer = UserOfferer(
            rights='admin',
            offerer=offerer,
            user=user
        )
        ApiHandler.save(user_offerer)

        # When
        user_offerer = UserOfferer.create_or_modify({
            '__SEARCH_BY__': ['offererId', 'userId'],
            'offererId': humanize(offerer.id),
            'rights': 'editor',
            'userId': humanize(user.id)
        })

        # Then
        assert user_offerer.offererId == offerer.id
        assert user_offerer.rights == 'editor'
        assert user_offerer.userId == user.id

    @with_delete
    def test_create_or_modify_returns_created_tag_with_nested_scope(self, app):
        # Given
        tag = Tag.create_or_modify({
            '__SEARCH_BY__': 'label',
            'label': 'Very High',
            'scopes': [
                {
                    'type': ScopeType.REVIEW
                }
            ]
        })

        # When
        ApiHandler.save(tag)

        # Then
        assert tag.scopes[0].tagId == tag.id

    @with_delete
    def test_create_or_modify_returns_modified_tag_with_nested_scope(self, app):
        # Given
        tag_dict = {
            '__SEARCH_BY__': 'label',
            'label': 'Very High',
            'scopes': [
                {
                    '__SEARCH_BY__': ['type'],
                    'type': ScopeType.REVIEW
                }
            ]
        }
        tag1 = Tag(**tag_dict)
        ApiHandler.save(tag1)

        # When
        tag2 = Tag.create_or_modify(tag_dict)
        ApiHandler.save(tag2)

        # Then
        assert tag2.id == tag1.id
        assert len(tag2.scopes) == 1
        assert tag2.scopes[0].tagId == tag2.id

    @with_delete
    def test_foo(self, app):
        # Given
        TAGS = [
            {
                'label': 'Accurate',
            },
            {
                'label': 'Biased',
            }
        ]
        for tag in TAGS:
            tag.update({
                '__SEARCH_BY__': ['label', 'type'],
                'id': '__NEXT_ID_IF_NOT_EXISTS__',
                'scopes': [{
                    '__SEARCH_BY__': ['tagId', 'type'],
                    'tagId': {
                        'humanized': True,
                        'key': 'id',
                        'type': '__PARENT__'
                    },
                    'type': ScopeType.REVIEW
                }],
                'type': TagType.QUALIFICATION
            })

        tags1 = []
        for tag_dict in TAGS:
            tag = Tag.create_or_modify(tag_dict)
            tags1.append(tag)
        ApiHandler.save(*tags1)

        # When
        tags2 = []
        for tag_dict in TAGS:
            tag = Tag.create_or_modify(tag_dict)
            tags2.append(tag)
        ApiHandler.save(*tags2)

        assert '/'.join([str(tag.id) for tag in tags1]) == '/'.join([str(tag.id) for tag in tags2])
        assert '/'.join([str(tag.scopes[0].id) for tag in tags1]) == '/'.join([str(tag.scopes[0].id) for tag in tags2])



import pytest
from datetime import datetime
from uuid import uuid4
from flask_login import login_user
from sqlalchemy import Integer
from sqlalchemy_api_handler import ApiHandler, humanize
from sqlalchemy_api_handler.serialization import as_dict

from tests.conftest import with_delete
from api.models.activity import Activity
from api.models.offer import Offer
from api.models.offer import Stock
from api.models.user import User


class ActivatorTest:
    @with_delete
    def test_create_offer_saves_an_insert_activity(self, app):
        # Given
        offer_dict = { 'name': 'bar', 'type': 'foo' }
        offer = Offer(**offer_dict)

        # When
        ApiHandler.save(offer)

        # Then
        activity = Activity.query.filter(Activity.changed_data['id'].astext.cast(Integer) == offer.id).one()
        assert activity.verb == 'insert'
        assert activity.oldDatum == None
        assert activity.transaction == None
        assert offer_dict.items() <= activity.patch.items()
        assert offer_dict.items() <= activity.datum.items()
        assert activity.datum['id'] == humanize(offer.id)
        assert activity.patch['id'] == humanize(offer.id)


    @with_delete
    def test_create_offer_with_login_user_saves_an_insert_activity_with_transaction(self, app):
        # Given
        offer_dict = { 'name': 'bar', 'type': 'foo' }
        offer = Offer(**offer_dict)
        user = User(email='fee@bar.com', firstName='fee', publicName='foo')
        ApiHandler.save(user)
        login_user(user)

        # When
        ApiHandler.save(offer)

        # Then
        activity = Activity.query.filter(Activity.changed_data['id'].astext.cast(Integer) == offer.id).one()
        assert activity.transaction.actor.id == user.id

    @with_delete
    def test_create_offer_saves_an_insert_activity(self, app):
        # Given
        offer_dict = { 'name': 'bar', 'type': 'foo' }
        offer = Offer(**offer_dict)
        ApiHandler.save(offer)
        modify_dict = { 'name': 'bor' }
        offer.modify(modify_dict)

        # When
        ApiHandler.save(offer)

        # Then
        activity = Activity.query.filter(
            (Activity.tableName == 'offer') &
            (Activity.verb == 'update') &
            (Activity.data['id'].astext.cast(Integer) == offer.id)
        ).one()
        assert {**offer_dict, **modify_dict}.items() <= activity.datum.items()
        assert modify_dict.items() == activity.patch.items()
        assert offer_dict.items() <= activity.oldDatum.items()

    @with_delete
    def test_create_activity_offer_saves_an_insert_activity(self, app):
        # Given
        offer_uuid = uuid4()
        patch = { 'name': 'bar', 'type': 'foo' }
        activity = Activity(dateCreated=datetime.utcnow(),
                            patch=patch,
                            tableName='offer',
                            uuid=offer_uuid)

        # When
        ApiHandler.activate(activity)

        # Then
        activity = Activity.query.filter_by(uuid=offer_uuid).one()
        offer = Offer.query.filter_by(activityUuid=offer_uuid).one()
        assert activity.verb == 'insert'
        assert patch.items() <= activity.datum.items()
        assert patch.items() <= activity.patch.items()
        assert activity.datum['id'] == humanize(offer.id)
        assert activity.patch['id'] == humanize(offer.id)

    @with_delete
    def test_create_two_activities_offer_saves_an_insert_and_an_update_activity(self, app):
        # Given
        offer_uuid = uuid4()
        first_patch = { 'name': 'bar', 'type': 'foo' }
        first_activity = Activity(dateCreated=datetime.utcnow(),
                                  patch=first_patch,
                                  tableName='offer',
                                  uuid=offer_uuid)
        second_patch = { 'name': 'bor' }
        second_activity = Activity(dateCreated=datetime.utcnow(),
                                   patch=second_patch,
                                   tableName='offer',
                                   uuid=offer_uuid)

        # When
        ApiHandler.activate(first_activity, second_activity)

        # Then
        activities = Activity.query.filter_by(uuid=offer_uuid).all()
        offer = Offer.query.filter_by(activityUuid=offer_uuid).one()
        assert len(activities) == 2
        assert offer.name == 'bor'


    @with_delete
    def test_create_activity_stock_binds_relationship_with_offer(self, app):
        # Given
        offer_uuid = uuid4()
        offer_patch = { 'name': 'bar', 'type': 'foo' }
        offer_activity = Activity(dateCreated=datetime.utcnow(),
                                  patch=offer_patch,
                                  tableName='offer',
                                  uuid=offer_uuid)
        stock_uuid = uuid4()
        stock_patch = { 'offerActivityUuid': offer_uuid, 'price': 3 }
        stock_activity = Activity(dateCreated=datetime.utcnow(),
                                  patch=stock_patch,
                                  tableName='stock',
                                  uuid=stock_uuid)

        # When
        ApiHandler.activate(offer_activity, stock_activity)

        # Then
        offer = Offer.query.filter_by(activityUuid=offer_uuid).one()
        stock = Stock.query.filter_by(activityUuid=stock_uuid).one()
        assert stock.offerId == offer.id



# https://medium.com/hackernoon/how-to-run-asynchronous-web-requests-in-parallel-with-python-3-5-without-aiohttp-264dc0f8546
from time import sleep
from functools import partial
import asyncio
from concurrent.futures import ThreadPoolExecutor


def chunks_from(elements, chunk_by=None):
    length = chunk_by if chunk_by else len(elements)
    for index in range(0, len(elements), length):
        yield elements[index:index + length]


def create_asynchronous(func,
                        args_list=None,
                        kwargs_list=None,
                        max_workers=10):
    if args_list and kwargs_list:
        if len(args_list) != len(kwargs_list):
            raise 'args_list and kwargs_list must have the same length.'
    elif args_list and not kwargs_list:
        kwargs_list = [{}]*len(args_list)
    elif kwargs_list:
        args_list = [()]*len(kwargs_list)

    async def asynchronous_func():
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            loop = asyncio.get_event_loop()
            tasks = [
                loop.run_in_executor(executor, partial(func, **kwargs), *args)
                for (args, kwargs) in zip(args_list, kwargs_list)
            ]
            return await asyncio.gather(*tasks)
    return asynchronous_func


def async_map(func,
              args_list=None,
              kwargs_list=None,
              chunk_by=None,
              max_workers=None,
              sleep_between=None):
    if args_list and kwargs_list:
        listed_args_list = list(args_list)
        listed_kwargs_list = list(kwargs_list)
        if len(listed_args_list) != len(listed_kwargs_list):
            raise 'args_list and kwargs_list must have the same length.'
        args_chunks = list(chunks_from(listed_args_list, chunk_by))
        kwargs_chunks = list(chunks_from(listed_kwargs_list, chunk_by))
    elif args_list and not kwargs_list:
        listed_args_list = list(args_list)
        if len(listed_args_list) == 0:
            return []
        args_chunks = list(chunks_from(listed_args_list, chunk_by))
        kwargs_chunks = None
    elif kwargs_list:
        listed_kwargs_list = list(kwargs_list)
        if len(listed_kwargs_list) == 0:
            return []
        args_chunks = None
        kwargs_chunks = list(chunks_from(listed_kwargs_list, chunk_by))
    elif args_list is None and kwargs_list is None:
        raise 'one args_list or kwargs_list need to be specified.'
    else:
        return []

    results = []
    for index in range(0, len(args_chunks)):
        asynchronous_func = create_asynchronous(func,
                                                args_list=args_chunks[index] if args_chunks else None,
                                                kwargs_list=kwargs_chunks[index] if kwargs_chunks else None,
                                                max_workers=max_workers)
        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(asynchronous_func())
        loop.run_until_complete(future)
        results += future.result()
        if sleep_between:
            sleep(sleep_between)
    return results



import inspect

from sqlalchemy_api_handler.bases.errors import NotSoftDeletableMixinException, \
                                                SoftDeletedRecordException


class SoftDelete():

    def __init__(self):
        self.isSoftDeleted = False

    def check_not_soft_deleted(self):
        if self.is_soft_deleted():
            api_errors = SoftDeletedRecordException()
            api_errors.add_error('check_not_soft_deleted', 'Entity already soft deleted')
            raise api_errors

    def is_soft_deleted(self):
        classes = inspect.getmro(type(self))
        if 'SoftDeletableMixin' in [cl.__name__ for cl in classes]:
            return self.isSoftDeleted
        return False

    @staticmethod
    def soft_delete(*entities):
        for entity in entities:
            classes = inspect.getmro(type(entity))
            if 'SoftDeletableMixin' not in [cl.__name__ for cl in classes]:
                api_errors = NotSoftDeletableMixinException()
                api_errors.add_error('soft_delete', 'cannot soft delete this entity')
                raise api_errors
            entity.isSoftDeleted = True



from logging import basicConfig, \
                    getLogger, \
                    CRITICAL, \
                    DEBUG, \
                    ERROR, \
                    INFO, \
                    WARNING, \
                    log
import os

from sqlalchemy_api_handler.utils.attr_dict import AttrDict

LOG_LEVEL = int(os.environ.get('LOG_LEVEL', INFO))

basicConfig(format='%(asctime)s %(levelname)-8s %(message)s',
                    level=LOG_LEVEL,
                    datefmt='%Y-%m-%d %H:%M:%S')


# this is so that we can have log.debug(XXX) calls in the app
# without XXX being evaluated when not at debug level
# this allows args to log.debug & co. to be lambdas that will
# get called when the loglevel is right
# cf. datascience/offers, in which the data printed in
# debug calls is costly to compute.
def api_logging(level, *args):
    if getLogger().isEnabledFor(level):
        evaled_args = map(lambda a: a() if callable(a) else a,
                          args)
        log(level, *evaled_args)


logger = AttrDict()
logger.critical = lambda *args: api_logging(CRITICAL, *args)
logger.debug = lambda *args: api_logging(DEBUG, *args)
logger.error = lambda *args: api_logging(ERROR, *args)
logger.info = lambda *args: api_logging(INFO, *args)
logger.warning = lambda *args: api_logging(WARNING, *args)



from sqlalchemy import Column, \
                       desc
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy_api_handler.bases.activator import Activator
from sqlalchemy.orm.collections import InstrumentedList


class HasActivitiesMixin(object):
    __versioned__ = {}

    activityUuid = Column(UUID(as_uuid=True))

    @property
    def activities(self):
        Activity = Activator.get_activity()
        query_filter = (Activity.table_name == self.__tablename__) & \
                       (Activity.uuid == self.activityUuid)
        return InstrumentedList(Activity.query.filter(query_filter) \
                                              .order_by(desc(Activity.id)) \
                                              .all())



import binascii
from base64 import b32encode, b32decode

from sqlalchemy_api_handler.utils.is_id_column import is_id_column
# This library creates IDs for use in our URLs,
# trying to achieve a balance between having a short
# length and being usable by humans
# We use base32, but replace O and I, which can be mistaken for 0 and 1
# by 8 and 9

class NonDehumanizableId(Exception):
    pass


def dehumanize(publicId):
    """ Get back an integer from a human-compatible ID """
    if publicId is None:
        return None
    missing_padding = len(publicId) % 8
    if missing_padding != 0:
        publicId += '=' * (8 - missing_padding)
    try:
        xbytes = b32decode(publicId.replace('8', 'O').replace('9', 'I'))
    except binascii.Error:
        raise NonDehumanizableId('id non dehumanizable')
    return int_from_bytes(xbytes)


def int_from_bytes(xbytes):
    return int.from_bytes(xbytes, 'big')


def dehumanize_ids_in(datum, model):
    if not datum:
        return None
    dehumanized_datum = {**datum}
    for (key, value) in datum.items():
        if hasattr(model, key):
            if is_id_column(getattr(model, key)):
                dehumanized_datum[key] = dehumanize(value)
    return dehumanized_datum



import pytest
from sqlalchemy_api_handler.serialization import as_dict
from api.models.offer import Offer
from api.models.stock import Stock
from api.models.user import User

class AsDictTest:
    def test_simple_dictify(self, app):
        # given
        user_fields_dict = {
            'email': 'marx.foo@plop.fr',
            'firstName' : 'Marx',
            'lastName': 'Foo',
            'publicName': 'Marx Foo'
        }
        user = User(**user_fields_dict)

        # when
        user_dict = as_dict(user)

        # then
        assert set(user_fields_dict).issubset(user_dict)

    def test_dictify_with_removing_includes(self, app):
        # given
        user_fields_dict = {
            'email': 'marx.foo@plop.fr',
            'firstName' : 'Marx',
            'lastName': 'Foo',
            'publicName': 'Marx Foo'
        }
        user = User(**user_fields_dict)

        # when
        user_dict = as_dict(user, includes=["-email"])

        # then
        user_fields_dict_without_email = dict(user_fields_dict)
        del user_fields_dict_without_email['email']
        assert set(user_fields_dict_without_email).issubset(user_dict)
        assert 'email' not in user_dict

    def test_dictify_with_relationships_includes(self, app):
        # given
        offer = Offer()
        stock = Stock()
        offer.stocks = [stock]

        # when
        stock_dict = as_dict(stock)
        offer_dict = as_dict(offer, includes=["stocks"])

        # then
        assert 'stocks' in offer_dict
        assert len(offer_dict['stocks']) == 1
        assert offer_dict['stocks'][0]['id'] == stock_dict['id']

    def test_dictify_with_default_class_includes(self, app):
        # given
        user_fields_dict = {
            'email': 'marx.foo@plop.fr',
            'firstName' : 'Marx',
            'lastName': 'Foo',
            'metier': 'philosophe',
            'publicName': 'Marx Foo'
        }
        user = User(**user_fields_dict)

        # when
        user_dict = as_dict(user)

        # then
        assert 'metier' not in user_dict
        assert user_dict['job'] == user_fields_dict['metier']

    def test_dictify_with_only_includes(self, app):
        # given
        user_fields_dict = {
            'email': 'marx.foo@plop.fr',
            'firstName' : 'Marx',
            'lastName': 'Foo',
            'metier': 'philosophe',
            'publicName': 'Marx Foo'
        }
        user = User(**user_fields_dict)

        # when
        includes = ['email', 'metier']
        user_dict = as_dict(user,
                            includes=includes,
                            mode='only-includes')

        # then
        assert len(user_dict) == len(includes)
        assert set(user_dict.keys()) == set(includes)
        assert set(user_dict.values()) == set(map(lambda key: user_fields_dict[key], includes))



from sqlalchemy_api_handler.utils.dehumanize import dehumanize
from sqlalchemy_api_handler.utils.humanize import humanize
from sqlalchemy_api_handler.utils.is_id_column import is_id_column


def dehumanize_ids_in(datum, model):
    if not datum:
        return None
    dehumanized_datum = {**datum}
    for (key, value) in datum.items():
        if hasattr(model, key):
            if is_id_column(getattr(model, key)):
                dehumanized_datum[key] = dehumanize(value)
    return dehumanized_datum


def humanize_ids_in(datum, model):
    if not datum:
        return None
    humanized_datum = {**datum}
    for (key, value) in datum.items():
        if hasattr(model, key):
            if is_id_column(getattr(model, key)):
                humanized_datum[key] = humanize(value)
    return humanized_datum


def synonyms_in(datum, model):
    synonymized_datum = {**datum}
    for synonym in model.__mapper__.synonyms:
        column_key = synonym._proxied_property.columns[0].key
        if column_key in synonymized_datum:
            synonymized_datum[synonym.key] = synonymized_datum[column_key]
            del synonymized_datum[column_key]
    return synonymized_datum


def relationships_in(datum, model):
    relationed_datum = {**datum}
    for (key, relationship) in model.__mapper__.relationships.items():
        uuid_key = '{}ActivityUuid'.format(key)
        if uuid_key in relationed_datum:
            model = relationship.mapper.class_
            instance = model.query.filter_by(activityUuid=relationed_datum[uuid_key]) \
                                  .first()
            relationed_datum[key] = instance
    return relationed_datum



import inflect
from functools import reduce
from itertools import groupby
from sqlalchemy import BigInteger

from sqlalchemy_api_handler.api_errors import ApiErrors
from sqlalchemy_api_handler.bases.save import Save


def merged_datum_from_activities(activities, initial=None):
    return reduce(lambda agg, activity: {**agg, **activity.patch},
                  activities,
                  initial if initial else {})


class Activator(Save):
    @classmethod
    def get_activity(cls):
        return Activator.activity_cls

    @classmethod
    def set_activity(cls, activity_cls):
        Activator.activity_cls = activity_cls

    @staticmethod
    def activate(*activities):
        Activity = Activator.get_activity()
        for (uuid, grouped_activities) in groupby(activities, key=lambda activity: activity.uuid):
            grouped_activities = sorted(grouped_activities, key=lambda activity: activity.dateCreated)

            first_activity = grouped_activities[0]
            table_name = first_activity.tableName
            model = Save.model_from_table_name(table_name)
            if model is None:
                errors = ApiErrors()
                errors.add_error('activity', 'model from {} not found'.format(table_name))
                raise errors

            id_key = model.id.property.key
            entity_id = first_activity.old_data.get(id_key) \
                        if first_activity.old_data else None
            if not entity_id:
                entity = model(**first_activity.patch)
                entity.activityUuid = uuid
                Save.save(entity)
                insert_activity = Activity.query.filter(
                    (Activity.tableName == table_name) & \
                    (Activity.data[id_key].astext.cast(BigInteger) == entity.id) & \
                    (Activity.verb == 'insert')
                ).one()
                insert_activity.dateCreated = first_activity.dateCreated
                insert_activity.uuid = uuid
                Save.save(insert_activity)
                first_activity.id = insert_activity.id
                first_activity.changed_data = insert_activity.changed_data
                first_activity.transaction = insert_activity.transaction
                for activity in grouped_activities[1:]:
                    activity.old_data = { id_key: entity.id }
                Activator.activate(*grouped_activities[1:])
                continue

            min_date = min(map(lambda a: a.dateCreated, grouped_activities))
            already_activities_since_min_date = Activity.query \
                                                        .filter(
                                                            (Activity.tableName == table_name) & \
                                                            (Activity.data[id_key].astext.cast(BigInteger) == entity_id) & \
                                                            (Activity.dateCreated >= min_date)
                                                        ) \
                                                        .all()

            Save.save(*grouped_activities)
            all_activities_since_min_date = sorted(already_activities_since_min_date + grouped_activities,
                                                   key=lambda activity: activity.dateCreated)
            datum = merged_datum_from_activities(all_activities_since_min_date,
                                                 all_activities_since_min_date[0].datum)
            del datum[model.id.key]
            datum['activityUuid'] = uuid
            entity = model.query.get(entity_id)
            entity.modify(datum)
            Save.save(entity)



from sqlalchemy_api_handler.bases.add import Add


class Delete(Add):

    @staticmethod
    def delete(*entities):
        db = Add.get_db()
        for entity in entities:
             db.session.delete(entity)

        db.session.commit()



import pytest
from sqlalchemy_api_handler import ApiErrors, ApiHandler
from sqlalchemy_api_handler.serialization import get_result
from sqlalchemy_api_handler.utils import humanize

from tests.conftest import with_delete
from api.models.offer import Offer
from api.models.stock import Stock

class GetResultTest:
    @with_delete
    def test_return_only_not_soft_deleted_stocks(self, app):
        # Given
        offer = Offer(name="foo", type="ThingType.JEUX_ABO")
        ApiHandler.save(offer)
        stock1 = Stock(price=1)
        stock1.offer = offer
        stock2 = Stock(price=2)
        stock2.offer = offer
        stock3 = Stock(price=3)
        stock3.offer = offer
        stock4 = Stock(price=4)
        stock4.offer = offer
        stock1.isSoftDeleted = True
        ApiHandler.save(stock1, stock2, stock3, stock4)

        # When
        result = get_result(Stock)
        data = result['data']

        # Then
        assert data[0]['id'] == humanize(stock2.id)
        assert data[1]['id'] == humanize(stock3.id)
        assert data[2]['id'] == humanize(stock4.id)

    @with_delete
    def test_check_order_by(self, app):
        # When
        with pytest.raises(ApiErrors) as e:
            get_result(Stock, order_by='(SELECT * FROM "user")')

        # Then
        assert 'order_by' in e.value.errors

    @with_delete
    def test_returns_total_data_count_with_has_more(self, app):
        # Given
        offer = Offer(name="foo", type="ThingType.JEUX_ABO")
        ApiHandler.save(offer)
        page = 2
        paginate = 10
        prices_length = 35
        stocks = []
        for price in range(prices_length):
            stock = Stock(price=price)
            stock.offer = offer
            stocks.append(stock)
        ApiHandler.save(*stocks)

        # When
        result = get_result(
            Stock,
            page=page,
            paginate=paginate,
            with_total_data_count=True
        )

        # Then
        assert len(result['data']) == paginate
        assert result['has_more'] == True
        assert result['total_data_count'] == prices_length

    @with_delete
    def test_returns_total_data_count_with_has_more(self, app):
        # Given
        offer = Offer(name="foo", type="ThingType.JEUX_ABO")
        ApiHandler.save(offer)
        page = 4
        paginate = 10
        prices_length = 35
        stocks = []
        for price in range(prices_length):
            stock = Stock(price=price)
            stock.offer = offer
            stocks.append(stock)
        ApiHandler.save(*stocks)

        # When
        result = get_result(
            Stock,
            page=page,
            paginate=paginate,
            with_total_data_count=True
        )

        # Then
        assert len(result['data']) == page * paginate - prices_length
        assert result['has_more'] == False
        assert result['total_data_count'] == prices_length



from datetime import datetime

def match_format(value: str, format: str):
    try:
        datetime.strptime(value, format)
    except ValueError:
        return False
    else:
        return True


def deserialize_datetime(key, value):
    valid_patterns = ['%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%S', '%Y-%m-%dT%H:%M:%SZ']
    datetime_value = None

    for pattern in valid_patterns:
        if match_format(value, pattern):
            datetime_value = datetime.strptime(value, pattern)

    return datetime_value


def format_into_ISO_8601(value):
    return value.isoformat() + "Z"


class DateTimes:
    def __init__(self, *datetimes):
        self.datetimes = list(datetimes)

    def __eq__(self, other):
        return self.datetimes == other.datetimes



import io
import pathlib
import os
import re

from setuptools import setup


HERE = pathlib.Path(__file__).parent
INSTALL_REQUIRES = (HERE / 'requirements.txt').read_text().splitlines()
TESTS_REQUIRE = (HERE / 'requirements-test.txt').read_text().splitlines()[1:]

with io.open('README.md', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open("{}/__init__.py".format(os.environ.get('MODULE_NAME')), "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read(), re.M).group(1)

setup(name=os.environ.get('PIP_NAME'),
      version=version,
      url='https://github.com/{}'.format(os.environ.get('GITHUB_NAME')),
      project_urls={
        'Code': 'https://github.com/{}'.format(os.environ.get('GITHUB_NAME')),
        'Issue tracker': 'https://github.com/{}/issues'.format(os.environ.get('GITHUB_NAME')),
      },
      license='MPL-2.0',
      author='Erwan Ledoux',
      author_email='lawanledoux@gmail.com',
      maintainer='Ledoux',
      maintainer_email='lawanledoux@gmail.com',
      description='Adds SQLAlchemy support to your Flask application for handle apis.',
      long_description=readme,
      long_description_content_type='text/markdown',
      packages=[
        'sqlalchemy_api_handler',
        'sqlalchemy_api_handler.bases',
        'sqlalchemy_api_handler.mixins',
        'sqlalchemy_api_handler.serialization',
        'sqlalchemy_api_handler.utils'
      ],
      include_package_data=True,
      python_requires=">= 2.7, != 3.0.*, != 3.1.*, != 3.2.*, != 3.3.*",
      install_requires=INSTALL_REQUIRES,
      tests_require=TESTS_REQUIRE,
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
       ],
)



from sqlalchemy_api_handler.mixins.activity_mixin import *
from sqlalchemy_api_handler.mixins.has_activities_mixin import *
from sqlalchemy_api_handler.mixins.needs_validation_mixin import *
from sqlalchemy_api_handler.mixins.soft_deletable_mixin import *



from datetime import datetime, timedelta
import pytest
from sqlalchemy_api_handler.serialization import serialize

from api.models.offer import Offer, ThingType
from api.models.time_interval import TimeInterval
from api.models.stock import Stock

time_interval = TimeInterval()
time_interval.start = datetime(2018, 1, 1, 10, 20, 30, 111000)
time_interval.end = datetime(2018, 2, 2, 5, 15, 25, 222000)
now = datetime.utcnow()


class SerializeTest:
    def test_on_datetime_list_returns_string_with_date_in_ISO_8601_list(self):
        # Given
        offer = Offer()
        stock = Stock()
        stock.offer = offer
        stock.beginningDatetime = now
        stock.endDatetime = now + timedelta(hours=3)
        offer.stocks = [stock]

        # When
        serialized_list = serialize(offer.dateRange)

        # Then
        for date in serialized_list:
            self._assert_is_in_ISO_8601_format(date)

    def test_on_enum_returns_dict_with_enum_value(self):
        # Given
        enum = ThingType.JEUX

        # When
        serialized_enum = serialize(enum)

        # Then
        assert serialized_enum == {
            'conditionalFields': [],
            'proLabel': 'Jeux (support physique)',
            'appLabel': 'Support physique',
            'offlineOnly': True,
            'onlineOnly': False,
            'sublabel': 'Jouer',
            'description': 'Résoudre l’énigme d’un jeu de piste dans votre ville ? '
                           'Jouer en ligne entre amis ? '
                           'Découvrir cet univers étrange avec une manette ?',
            'isActive': False
        }

    def _assert_is_in_ISO_8601_format(self, date_text):
        try:
            format_string = '%Y-%m-%dT%H:%M:%S.%fZ'
            datetime.strptime(date_text, format_string)
        except TypeError:
            assert False, 'La date doit être un str'
        except ValueError:
            assert False, 'La date doit être au format ISO 8601 %Y-%m-%dT%H:%M:%S.%fZ'



import inflect
from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship, synonym

from sqlalchemy_api_handler.utils.datum import relationships_in, \
                                               synonyms_in
from sqlalchemy_api_handler.utils.dehumanize import dehumanize_ids_in
from sqlalchemy_api_handler.utils.humanize import humanize, humanize_ids_in


inflect_engine = inflect.engine()


class ActivityMixin(object):
    uuid = Column(UUID(as_uuid=True))

    @declared_attr
    def dateCreated(cls):
        return synonym('issued_at')

    @declared_attr
    def tableName(cls):
        return synonym('table_name')

    @property
    def collectionName(self):
        return inflect_engine.plural_noun(self.tableName)

    @property
    def datum(self):
        if not self.data:
            return None
        model = self.__class__.model_from_table_name(self.tableName)
        return relationships_in(synonyms_in(humanize_ids_in(self.data, model), model), model)

    @property
    def entity(self):
        model = self.__class__.model_from_table_name(self.tableName)
        return model(**self.datum)

    @property
    def oldDatum(self):
        if not self.old_data:
            return None
        model = self.__class__.model_from_table_name(self.tableName)
        return relationships_in(synonyms_in(humanize_ids_in(self.old_data, model), model), model)

    @property
    def patch(self):
        if not self.changed_data:
            return None
        model = self.__class__.model_from_table_name(self.tableName)
        return relationships_in(synonyms_in(humanize_ids_in(self.changed_data, model), model), model)

    def modify(self,
               datum,
               skipped_keys=[],
               with_add=False):
        dehumanized_datum = {**datum}
        model = self.__class__.model_from_table_name(datum.get('tableName', self.tableName))
        for (humanized_key, dehumanized_key) in [('oldDatum', 'old_data'), ('patch', 'changed_data')]:
            if humanized_key in dehumanized_datum:
                dehumanized_datum[dehumanized_key] = dehumanize_ids_in(dehumanized_datum[humanized_key],
                                                                       model)
                del dehumanized_datum[humanized_key]
        super().modify(dehumanized_datum,
                       skipped_keys=skipped_keys,
                       with_add=with_add)

    __as_dict_includes__ = [
        'collectionName',
        'dateCreated',
        'patch',
        'tableName',
        '-changed_data',
        '-issued_at',
        '-native_transaction_id',
        '-old_data',
        '-table_name',
        '-schema_name',
        '-transaction_id',
        '-verb'
    ]
